#!/usr/bin/perl
use POSIX;
use Errno;
use IO::Socket;
use IO::Select;
use Tie::RefHash ();

my $DaemonPort = 5190;			#-- ,    
my $min_login_length = 3;		#--   
my $shutdown = 0;
tie %logins, 'Tie::RefHash';	#-- $logins{$client} = login
tie %in_buf, 'Tie::RefHash';
my (%out_buf, $TIME);

close STDIN; close STDERR;
fork && exit;

$|=1;
open(LOG_FILE, ">> chat.log");

_log("Server started...");
my $server = IO::Socket::INET->new(LocalPort => $DaemonPort, Listen => 10 ) or die "Can't make server socket: $@n";
my $select = IO::Select->new($server);
while (! $shutdown) {
	#--  
	my ($sec,$min,$hour) = localtime(time);
	$TIME = sprintf("%02d:%02d", $hour,$min);

	#--    ,   .
	foreach my $client ($select->can_read(0.1)) {
		if($client == $server) { #--    .
			$client = $server->accept();
			$client->autoflush(1);
			$select->add($client);
		} else {	#--    .
			my $data = '';
			my $rv   = $client->recv($data, POSIX::BUFSIZ, 0);	#--  
			unless (defined($rv) && length $data) {		#--    ,    
				disconnect($client);
				next;
			}
			while (($_=index($data, chr(0))) >= 0) {	#--      chr(0)
				push(@{$in_buf{$client}}, substr($data, 0, $_, ''));	#--     %in_buf
				substr($data, 0, 1, '');
			}
		}
	}

	#--      
	foreach my $client (keys %in_buf) {
		foreach my $cmd (@{$in_buf{$client}}) {
			my ($out, $no_utf)=goCommand($client, $cmd);	#--   
			if ($out) {		#--     
				if ($no_utf) { $out_buf{$client} .= $out }	#--      utf
				else { $out_buf{$client} .= utf($out) }		#--     utf
			}
		}
	}
	%in_buf = ();

	#--     
	foreach my $client ($select->can_write(0.1)) {	#--    ?
		next unless exists $out_buf{$client};		#--    
		my $rv = $client->send("".$out_buf{$client}.chr(0), 0);	#--     chr(0)
		unless (defined $rv) {	#-- -  
			_log("I was told I could write, but I can't.");
			next;
		}
		if ($rv == length($out_buf{$client})+1 or Errno::EINPROGRESS == $! or Errno::EWOULDBLOCK ) {	#--    ,    
			substr($out_buf{$client}, 0, $rv) = '';
			delete $out_buf{$client} unless length($out_buf{$client});
		} else { #--    .  ? ...
			disconnect($client);
			next;
		}
	}
}
$server->close();
close LOG_FILE;
_log("Server shutdown");
exit;

#--     
sub _log
{   my ($s) = @_;
	print LOG_FILE "".(localtime(time))."\t$s\n";
}

#--  
sub disconnect
{	my $client = shift;
	_log("disconnect: $logins{$client}");
	my $login = $logins{$client};
	#--        
	delete $in_buf{$client};
	delete $out_buf{$client};
	delete $logins{$client};
	#--    
	$select->remove($client);
	close $client;
	#--       
	if (length($login)>=$min_login_length) {
		SendAll("<DEL>$login</DEL>");
	}
#	$shutdown = 1;
}

#--   __    
sub SendAll
{   my ($mes) = @_;
	foreach my $client (keys %logins) {
		$out_buf{$client} .= $mes
	}
}
#--     ,   
sub SendTo
{   my ($mes, $name) = @_;
	foreach my $client (keys %logins) {
		next unless ($logins{$client} eq $name);
		$out_buf{$client} .= $mes;
		last;
	}
}

#--     
sub goCommand
{   my ($client, $line) = @_;
_log("goCommand: $line");
	if (index($line, "<LOGIN")==0) {	#-- 
		if ($line=~/NAME=\"([^\"]+)\"/) {
			my $login = $1;
			#--  
			if ($_ = CheckLogin($login)) { return "<ERROR TEXT=\"$_\"/>" }	#-- -   ,   
			#-- ,      .
			foreach my $client (keys %logins) {
				if (mylc($logins{$client}) eq mylc($login)) {	#--    
					return "<ERROR TEXT=\"  ,  \"/>";
				}
			}
			#-- ,    . ...
			_log("LOGIN: $login");
			SendAll("<ADD>$login</ADD>");
			$logins{$client} = $login;
			return ("<OK /><LIST>".join(',', values %logins)."</LIST>",1);
		} 
	} elsif (index($line, "<T")==0 and length($logins{$client})>=$min_login_length) {	#--    ,   
		if ($line =~ /(<T[^>]*>)([^<]+<\/T>)/) {
			my ($start, $txt) = ($1, "$TIME \[$logins{$client}\] $2");
			if (index($txt, 'private [')>0) {	#--    ,     
				while ($txt=~/private \[([^\]]+)\]/g) {	#--  ,    
					SendTo("$start$txt", $1) unless ($1 eq $logins{$client});
				}
				$out_buf{$client} .= "$start$txt";	#--     
			} else {	#--     ,      
				SendAll("$start$txt");
			}
		}
	}
	return "";
}

#--    
sub CheckLogin
{   my ($login) = shift;
	if (length($login)<$min_login_length) {
		return "      $min_login_length ";
	}
	if ($login=~/([;,&<>\"\'\`\#\?\[\]\t\n\(\)\=])/) {	#--     
		return "   ";
	}
	if ($login=~/^ / or $login=~/ $/) { return "      "; }
	if ($login=~/  /) { return "      "; }
	if ($login=~//) { return "      Alt+0160"; }
	return '';	#--  OK
}

#--     utf (    )
sub utf
{   my $s = shift;
	$s=~s/([--ܨ])/win2utf($1)/eg;
	return $s;
}

sub win2utf
{	my $s = shift;
	my $c208 = chr(208);
	my $c209 = chr(209);
	if ( (ord($s)>=192) and (ord($s)<=239)) { return $c208.chr(ord($s)-48) }
	if ( (ord($s)>=240) and (ord($s)<=255)) { return $c209.chr(ord($s)-112) }
	if ($s=="") { return $c208.chr(149) }
	if ($s=="") { return $c208.chr(181) }
	if ($s=="") { return $c208.chr(172) }
	if ($s=="") { return $c208.chr(170) }
	if ($s=="") { return $c208.chr(140) }
	if ($s=="") { return $c208.chr(138) }
	return $s;
}

#==   
sub mylc
{ my $s = shift;
  $s =~ tr/ިQWERTYUIOPASDFGHJKLZXCVBNM/qwertyuiopasdfghjklzxcvbnm/;
  return $s;
}
